package jdk.internal.instrumentation;

import jdk.internal.org.objectweb.asm.Type;
import jdk.internal.org.objectweb.asm.Handle;
import java.util.Iterator;
import jdk.internal.org.objectweb.asm.commons.RemappingMethodAdapter;
import jdk.internal.org.objectweb.asm.tree.MethodNode;
import java.util.LinkedList;
import jdk.internal.org.objectweb.asm.commons.SimpleRemapper;
import jdk.internal.org.objectweb.asm.MethodVisitor;
import java.util.HashMap;
import java.util.Map;
import java.lang.reflect.Method;
import java.util.List;
import jdk.internal.org.objectweb.asm.tree.ClassNode;
import jdk.internal.org.objectweb.asm.ClassVisitor;

final class MethodMergeAdapter extends ClassVisitor
{
    private final ClassNode cn;
    private final List<Method> methodFilter;
    private final Map<String, String> typeMap;
    private final Logger logger;
    
    public MethodMergeAdapter(ClassVisitor classVisitor, ClassNode cn, List<Method> methodFilter, TypeMapping[] array, Logger logger) {
        super(327680, classVisitor);
        this.cn = cn;
        this.methodFilter = methodFilter;
        this.logger = logger;
        this.typeMap = new HashMap();
        for (TypeMapping typeMapping : array) {
            this.typeMap.put(typeMapping.from().replace('.', '/'), typeMapping.to().replace('.', '/'));
        }
    }
    
    @Override
    public void visit(int version, int access, String name, String signature, String superName, String[] interfaces) {
        super.visit(version, access, name, signature, superName, interfaces);
        this.typeMap.put(this.cn.name, name);
    }
    
    @Override
    public MethodVisitor visitMethod(int access, String s, String s2, String signature, String[] exceptions) {
        if (this.methodInFilter(s, s2)) {
            this.logger.trace("Deleting " + s + s2);
            return null;
        }
        return super.visitMethod(access, s, s2, signature, exceptions);
    }
    
    @Override
    public void visitEnd() {
        SimpleRemapper remapper = new SimpleRemapper(this.typeMap);
        LinkedList<MethodNode> list = new LinkedList();
        for (MethodNode methodNode : this.cn.methods) {
            if (this.methodInFilter(methodNode.name, methodNode.desc)) {
                list.add(methodNode);
            }
            else {
                continue;
            }
        }
        while (!list.isEmpty()) {
            MethodNode methodNode2 = (MethodNode)list.remove(0);
            this.logger.trace("Copying method: " + methodNode2.name + methodNode2.desc);
            this.logger.trace("   with mapper: " + this.typeMap);
            MethodVisitor visitMethod = this.cv.visitMethod(methodNode2.access, methodNode2.name, methodNode2.desc, methodNode2.signature, (String[])methodNode2.exceptions.toArray(new String[0]));
            methodNode2.instructions.resetLabels();
            methodNode2.accept(new RemappingMethodAdapter(methodNode2.access, methodNode2.desc, visitMethod, remapper));
            this.findMethodsReferencedByInvokeDynamic(methodNode2, list);
        }
        super.visitEnd();
    }
    
    private void findMethodsReferencedByInvokeDynamic(final MethodNode methodNode, final List<MethodNode> list) {
        methodNode.accept(new MethodVisitor(327680) {
            @Override
            public void visitInvokeDynamicInsn(String s, String s2, Handle handle, Object... array) {
                for (Object o : array) {
                    if ((o instanceof Handle)) {
                        Handle handle2 = (Handle)o;
                        MethodNode access$100 = MethodMergeAdapter.access$100(MethodMergeAdapter.this.cn, handle2);
                        if (access$100 == null) {
                            MethodMergeAdapter.this.logger.error("Could not find method " + handle2.getName() + handle2.getDesc() + " referenced from an invokedynamic in " + methodNode.name + methodNode.desc + " while processing class " + MethodMergeAdapter.this.cn.name);
                        }
                        MethodMergeAdapter.this.logger.trace("Adding method referenced from invokedynamic " + access$100.name + access$100.desc + " to the list of methods to be copied from " + MethodMergeAdapter.this.cn.name);
                        list.add(access$100);
                    }
                }
            }
        });
    }
    
    private static MethodNode findMethod(ClassNode classNode, Handle handle) {
        for (MethodNode methodNode : classNode.methods) {
            if (methodNode.name.equals(handle.getName()) && methodNode.desc.equals(handle.getDesc())) {
                return methodNode;
            }
            continue;
        }
        return null;
    }
    
    private boolean methodInFilter(String anObject, String anObject2) {
        for (Method method : this.methodFilter) {
            if (method.getName().equals(anObject) && Type.getMethodDescriptor(method).equals(anObject2)) {
                return true;
            }
            continue;
        }
        return false;
    }
    
    static MethodNode access$100(ClassNode classNode, Handle handle) {
        return findMethod(classNode, handle);
    }
}
